###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Initialize the server based on environment variables."""
import json
import logging
import os
from importlib import resources
from importlib.metadata import entry_points
from typing import Callable

from django.db.models import Count
from django.db.utils import OperationalError

from .. import entrypoint_manager as epm
from .. import models as m
from ..plugin_manager import PluginManager

logger = logging.getLogger('ocr.general')

def run_on_env(env_name: str, func_map: dict[str, Callable]):
    """Run a function if the environment variable is set."""
    if env_name in os.environ:
        value = os.environ.get(env_name).lower()
        for key, func in func_map.items():
            if isinstance(key, str):
                key = key.lower()
                if key.lower() == value:
                    break
            elif isinstance(key, tuple):
                if any(k.lower() == value for k in key):
                    break
            else:
                raise ValueError(f'Invalid use of `run_on_env`: key `{key}` is not a string or tuple')
        else:
            logger.warning('Unknown value for environment variable `{env_name}`: {value}... Doing nothing')
            func = None

        if func is None:
            return

        try:
            func()
            logger.info(f'Ran `{func.__name__}` based on environment variable `{env_name}`')
        except OperationalError:
            msg = f'Ignoring environment variable `{env_name}` as the database is not ready/migrated.'
            logger.warning(msg)

def init_most_used():
    """Initialize the server with the most used languages and models."""
    logger.info('Initializing server with most used languages and models')
    src = m.Language.objects.annotate(count=Count('trans_src')).order_by('-count').first()
    dst = m.Language.objects.annotate(count=Count('trans_dst')).order_by('-count').first()

    if src and src.count > 0:
        m.Language.load_model_src(src.iso1)
    if dst and dst.count > 0:
        m.Language.load_model_dst(dst.iso1)

    box = m.OCRBoxModel.objects.annotate(count=Count('box_runs')).order_by('-count').first()
    ocr = m.OCRModel.objects.annotate(count=Count('ocr_runs')).order_by('-count').first()
    tsl = m.TSLModel.objects.annotate(count=Count('tsl_runs')).order_by('-count').first()

    if box and box.count > 0:
        m.OCRBoxModel.load_model(box.name)
    if ocr and ocr.count > 0:
        m.OCRModel.load_model(ocr.name)
    if tsl and tsl.count > 0:
        m.TSLModel.load_model(tsl.name)

def init_last_used():
    """Initialize the server with the most used languages and models."""
    logger.info('Initializing server with the last used languages and models')
    src = m.Language.get_last_loaded_src()
    dst = m.Language.get_last_loaded_dst()

    box = m.OCRBoxModel.get_last_loaded()
    ocr = m.OCRModel.get_last_loaded()
    tsl = m.TSLModel.get_last_loaded()

    if src:
        m.Language.load_model_src(src.iso1)
    else:
        logger.warning('No last source language found')
    if dst:
        m.Language.load_model_dst(dst.iso1)
    else:
        logger.warning('No last destination language found')
    if box:
        m.OCRBoxModel.load_model(box.name)
    else:
        logger.warning('No last box model found')
    if ocr:
        m.OCRModel.load_model(ocr.name)
    else:
        logger.warning('No last OCR model found')
    if tsl:
        m.TSLModel.load_model(tsl.name)
    else:
        logger.warning('No last TSL model found')

def auto_create_languages():
    """Create Language objects from json file."""
    lang_file = resources.files('ocr_translate.ocr_tsl').joinpath('languages.json')
    with lang_file.open(encoding='utf-8') as f:
        langs = json.load(f)

    for lang in langs:
        m.Language.from_dct(lang)

    m.OptionDict.objects.get_or_create(options={})

def load_ept_data(namespace):
    """Load all entrypoints from a namespace into a list"""
    return[_.load() for _ in entry_points(group=namespace)]

def ensure_plugins():
    """Ensure that all plugins are installed on initialization.
    This is used to make sure that running the server with a new DEVICE will have the correct dependencies installed."""
    #pylint: disable=import-outside-toplevel,cyclic-import
    from ..entrypoint_manager import ep_manager
    logger.info('Ensuring that all plugins are loaded')
    pmng = PluginManager()
    known = set(_['name'] for _ in pmng.plugins_data)
    installed = set(pmng.managed_plugins)
    for plugin in known & installed:
        with ep_manager():
            pmng.install_plugin(plugin)

def sync_models_epts():
    """Deactivate models that are in the database but whose entrypoint is no longer available."""
    for group, cls in epm.GROUPS.items():
        ept_data = load_ept_data(group)
        names_ep = set(_['name'] for _ in ept_data)

        # Deactivate models that do not have an entrypoint anymore
        q = cls.objects
        names_db_on = set(q.filter(active=True).values_list('name', flat=True))
        removed = names_db_on - names_ep
        for name in removed:
            model = cls.objects.get(name=name)
            model.deactivate()
            logger.info(f'Deactivated {cls.__name__:>12s} `{name}` as its entrypoint is no longer available')

        for model_data in ept_data:
            cls.from_dct(model_data)

def deprecate_los_true():
    """Deprecate the environment variable `LOAD_ON_START=true`."""
    logger.warning('The environment variable `LOAD_ON_START=true` is deprecated (defaults to `most`).')
    logger.warning('Use `LOAD_ON_START=most` or `LOAD_ON_START=last` instead.')
    init_most_used()

TRUE_VALUES = ('true', 't', '1')
FALSE_VALUES = ('false', 'f', '0')
RUN_ON_ENV_INIT = {
    'LOAD_ON_START': {
        'most': init_most_used,
        'last': init_last_used,
        TRUE_VALUES: deprecate_los_true,
        FALSE_VALUES: None
    }
}

def env_var_init():
    """Run initializations based on environment variables."""
    for env, func_map in RUN_ON_ENV_INIT.items():
        run_on_env(env, func_map)



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""
URL configuration for ocr_translate project.

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/4.2/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.contrib import admin
from django.urls import include, path

urlpatterns = [
    path('', include('ocr_translate.urls')),
    path('admin/', admin.site.urls),
]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""
Django settings for ocr_translate project.

Generated by 'django-admin startproject' using Django 4.2.2.

For more information on this file, see
https://docs.djangoproject.com/en/4.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.2/ref/settings/
"""
# pylint: skip-file
import json
import os
from pathlib import Path

from ocr_translate.plugin_manager import PluginManager

PMNG = PluginManager()

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(os.getenv('OCT_BASE_DIR', ''))
PROJECT_DIR = Path(__file__).resolve().parent.parent

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get('DJANGO_DEBUG', '').lower() in ['true', 't', '1', 'yes', 'y']
DJANGO_LOG_LEVEL = os.environ.get('DJANGO_LOG_LEVEL', 'INFO').upper()

STREAM_HANDLER = 'logging.StreamHandler'
STREAM_HANDLER_KWARGS = {}
MEDIUM_FMT_STR = '{asctime} - {levelname:>7s} - {name:>15s}:{module:<15s} - {message}'

try:
    from rich.logging import RichHandler
except:
    pass
else:
    STREAM_HANDLER = 'rich.logging.RichHandler'
    STREAM_HANDLER_KWARGS = {
        'rich_tracebacks': True,
        'tracebacks_suppress': ['django', 'logging', 'rich'],
    }
    MEDIUM_FMT_STR = '{message}'

LOGFILE = os.environ.get('OCT_LOGFILE', '')

handlers_list = ['console']
file_handler = {}
if LOGFILE:
    if LOGFILE.lower() in ['true', 't', '1', 'yes', 'y']:
        LOGFILE = BASE_DIR / 'ocr_translate.log'
    file_handler = {
        'file': {
            'level': 'DEBUG',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': LOGFILE.as_posix(),
            'formatter': 'verbose',
            'maxBytes': 10*1024*1024,  # 10 MB
            'backupCount': 5,
            'encoding': 'utf8',
        }
    }
    handlers_list.append('file')

def skip_static_requests(record):
    return not record.args[0].startswith('GET /static/')

# Logging
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '{levelname} - {asctime} - {module}:{funcName}:{lineno} - {process:d} - {thread:d} - {message}',
            'style': '{',
            },
        'medium': {
            'format': MEDIUM_FMT_STR,
            'style': '{',
            },
        'simple': {
            'format': '{levelname} - {message}',
            'style': '{',
            },
    },
    'filters': {
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
        'skip_static_requests': {
            '()': 'django.utils.log.CallbackFilter',
            'callback': skip_static_requests,
        }
    },
    'handlers': {
        'console': {
            'level': DJANGO_LOG_LEVEL,
            'class': STREAM_HANDLER,
            'formatter': 'medium',
            **STREAM_HANDLER_KWARGS,
        },
        **file_handler
    },
    'loggers': {
        'django': {
            'handlers': ['console'],
            'propagate': False,
        },
        'django.request': {
            'handlers': ['console'],
            'level': 'ERROR',
            'filters': ['skip_static_requests'],
            'propagate': False,
        },
        'ocr.general': {
            'handlers': handlers_list,
            'level': 'DEBUG',
            'propagate': False,
        },
        'ocr.worker': {
            'handlers': handlers_list,
            'level': 'DEBUG',
            'propagate': False,
        },
        'plugin': {
            'handlers': handlers_list,
            'level': 'DEBUG',
            'propagate': False,
        },
    },
}

def parse_list(s, sep=';'):
    if s.startswith('"') and s.endswith('"'):
        s = s[1:-1]
    if s.startswith("'") and s.endswith("'"):
        s = s[1:-1]
    return list(filter(None, s.split(sep)))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('DJANGO_SECRET_KEY', 'django-insecure-7h+*^e963rdi*2jbdlhqvmg%xnx$9@s*ccgcfae@t219$#!)vu')

ALLOWED_HOSTS = ['localhost', '127.0.0.1']
ALLOWED_HOSTS += parse_list(os.environ.get('DJANGO_ALLOWED_HOSTS', ''))

###################################################################################
# CORS
USE_CORS_HEADERS = os.environ.get('USE_CORS_HEADERS', 'false').lower() in ['true', 't', '1', 'yes', 'y']
CORS_ALLOWED_ORIGINS = parse_list(os.environ.get('CORS_ALLOWED_ORIGINS', ''))
CSRF_TRUSTED_ORIGINS = CORS_ALLOWED_ORIGINS
if 'CORS_ALLOW_METHODS' in os.environ:
    CORS_ALLOW_METHODS = parse_list(os.environ.get('CORS_ALLOW_METHODS', ''))
if 'CORS_ALLOW_HEADERS' in os.environ:
    CORS_ALLOW_HEADERS = parse_list(os.environ.get('CORS_ALLOW_HEADERS', ''))

# Application definition
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]
if USE_CORS_HEADERS:
    INSTALLED_APPS.append('corsheaders')
INSTALLED_APPS.append('ocr_translate')
INSTALLED_APPS += PMNG.all_plugins

# Middleware
MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
]
if USE_CORS_HEADERS:
    MIDDLEWARE.append('corsheaders.middleware.CorsMiddleware')
MIDDLEWARE += [
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'ocr_translate.app.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'ocr_translate.app.wsgi.application'


# Database
# https://docs.djangoproject.com/en/4.2/ref/settings/#databases

DATABASE_ENGINE = os.environ.get('DATABASE_ENGINE', 'django.db.backends.sqlite3')
DATABASE_NAME = os.environ.get('DATABASE_NAME', os.path.join(PROJECT_DIR, 'db.sqlite3'))

DATABASE = {
    'ENGINE': DATABASE_ENGINE,
    'NAME': DATABASE_NAME,
}

for k in ['USER', 'PASSWORD', 'HOST', 'PORT']:
    if not os.environ.get('DATABASE_' + k) is None:
        DATABASE[k] = os.environ.get('DATABASE_' + k)

OPTIONS = {}
if DATABASE_ENGINE == 'django.db.backends.mysql':
    # https://stackoverflow.com/questions/2108824/mysql-incorrect-string-value-error-when-save-unicode-string-in-django
    OPTIONS['charset'] = 'utf8mb4'

if DATABASE_ENGINE == 'django.db.backends.sqlite3':
    ptr = {}
    DATABASE['OPTIONS'] = ptr
    ptr['init_command'] = 'PRAGMA journal_mode=wal;'

DATABASES = {
    'default': DATABASE
}


# Password validation
# https://docs.djangoproject.com/en/4.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True

# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'
DATA_UPLOAD_MAX_MEMORY_SIZE = 10*1024*1024

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.2/howto/static-files/

STATIC_URL = '/static/'

STATICFILES_DIRS = (
    os.path.join(PROJECT_DIR, 'static'),
    )
STATIC_ROOT = os.path.join(PROJECT_DIR, 'staticfiles')
# MEDIA_URL = '/media/'
# MEDIA_ROOT = os.path.join(PROJECT_DIR, 'media')



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
import json
import logging
from importlib import resources
from importlib.metadata import entry_points
from typing import Type

from django.db import models

from ..ocr_tsl.signals import refresh_model_cache_signal
from ..trie import Trie

logger = logging.getLogger('ocr.general')

def safe_get_or_create(model: Type[models.Model], strict: bool = False, **kwargs) -> models.Model:
    """Get or create a model instance in a safe way.

    Args:
        model (Type[models.Model]): The model class to get or create.
        strict (bool, optional): Whether to raise a caught exception if ocurred. Defaults to False.
    """
    try:
        obj, _ = model.objects.get_or_create(**kwargs)
    except model.MultipleObjectsReturned as exc:
        if strict:
            raise exc
        logger.warning(f'Multiple objects returned for {model}: {model.objects.filter(**kwargs).all()}')
        obj = model.objects.filter(**kwargs).first()
    return obj

class OptionDict(models.Model):
    """Dictionary of options for OCR and translation"""
    options = models.JSONField(unique=True)

    def __str__(self):
        return str(self.options)

class LoadEvent(models.Model):
    """Event log for the OCR and translation tasks"""
    description = models.CharField(max_length=512)

    date = models.DateTimeField(auto_now_add=True)

class Image(models.Model):
    """Image registered as the md5 of the uploaded file"""
    md5 = models.CharField(max_length=32, unique=True)


class Text(models.Model):
    """Text extracted from an image or translated from another text"""
    text = models.TextField()
    # lang = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='texts')

class Language(models.Model):
    """Language used for translation"""
    LOADED_SRC: 'Language' = None
    LOADED_DST: 'Language' = None
    LOADED_TRIE: Trie = None

    name = models.CharField(max_length=64, unique=True)
    iso1 = models.CharField(max_length=8, unique=True)
    iso2b = models.CharField(max_length=8, unique=True)
    iso2t = models.CharField(max_length=8, unique=True)
    iso3 = models.CharField(max_length=32, unique=True)

    default_options = models.ForeignKey(
        OptionDict, on_delete=models.CASCADE, related_name='lang_default_options', null=True
        )

    load_events_src = models.ManyToManyField('LoadEvent', related_name='languages_src')
    load_events_dst = models.ManyToManyField('LoadEvent', related_name='languages_dst')

    def __str__(self):
        return f'{self.name} ({self.iso1})'

    def __eq__(self, other):
        if isinstance(other, Language):
            return self.iso1 == other.iso1
        if isinstance(other, str):
            return self.iso1 == other
        return False

    # https://stackoverflow.com/questions/61212514/django-model-objects-became-not-hashable-after-upgrading-to-django-2-2
    def __hash__(self):
        return hash(self.iso1)

    def load_src(self) -> None:
        """Load the language trie"""
        self.load_events_src.create(description=f'Loading SRC Language {self.name}')

    def load_dst(self) -> None:
        """Load the language trie"""
        self.load_events_dst.create(description=f'Loading DST Language {self.name}')

    @classmethod
    def from_dct(cls, data: dict) -> 'Language':
        """Create or update a language from a dictionary"""
        logger.debug(f'Creating/Updating Language from dict: {data}')

        data = data.copy()

        def_opt = data.pop('default_options', {})

        required = {'name', 'iso1', 'iso2t', 'iso2b', 'iso3'}
        missing = required - set(data.keys())
        if missing:
            raise ValueError(f'Missing required keys {missing} in data for Language')

        obj, _ = cls.objects.get_or_create(**data)
        opt_obj, created = OptionDict.objects.get_or_create(options=def_opt)
        if created:
            logger.debug(f'Created new OptionDict for Language {obj.name}: {def_opt}')
        obj.default_options = opt_obj
        obj.save()
        return obj

    @classmethod
    def get_last_loaded_src(cls) -> 'Language':
        """Get the last loaded language"""
        res = cls.objects
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events_src'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events_src__date')
        res = res.first()
        if res is None or len(res.load_events_src.all()) == 0:
            logger.debug('No load events found for Language')
            return None
        return res

    @classmethod
    def get_last_loaded_dst(cls) -> 'Language':
        """Get the last loaded language"""
        res = cls.objects
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events_dst'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events_dst__date')
        res = res.first()
        if res is None or len(res.load_events_dst.all()) == 0:
            logger.debug('No load events found for Language')
            return None
        return res

    @classmethod
    def load_trie(cls, iso1: str) -> Trie:
        """Load the language trie"""
        trie_file = resources.files('ocr_translate.dictionaries').joinpath(f'{iso1}.txt')
        freq_file = resources.files('ocr_translate.dictionaries').joinpath(f'{iso1}_freq.json')

        if not trie_file.exists():
            logger.debug(f'No source language trie found: {iso1}')
            cls.LOADED_TRIE = None
            return None

        with freq_file.open(encoding='utf-8') as f:
            freq = json.load(f)

        logger.info(f'Loading source language trie: {iso1}')
        res = Trie()
        with trie_file.open(encoding='utf-8') as f:
            for word in f.read().splitlines():
                res.insert(word, freq.get(word, -1e-4))

        cls.LOADED_TRIE = res
        return res

    @classmethod
    def load_model_src(cls, lang_iso1: str) -> 'Language':
        """Load a language by iso1 code and unload the current one if needed"""
        current = cls.LOADED_SRC
        if current is not None:
            if current.iso1 == lang_iso1:
                return current

        logger.info(f'Loading Language SRC model: {lang_iso1}')
        obj = cls.objects.get(iso1=lang_iso1)
        obj.load_events_src.create(description=f'Loading SRC Language {obj.name}')
        obj.load_trie(obj.iso1)

        cls.LOADED_SRC = obj
        refresh_model_cache_signal.send(sender=None)
        return obj

    @classmethod
    def load_model_dst(cls, lang_iso1: str) -> 'Language':
        """Load a language by iso1 code and unload the current one if needed"""
        current = cls.LOADED_DST
        if current is not None:
            if current.iso1 == lang_iso1:
                return current

        logger.info(f'Loading Language DST model: {lang_iso1}')
        obj = cls.objects.get(iso1=lang_iso1)
        obj.load_events_dst.create(description=f'Loading DST Language {obj.name}')

        cls.LOADED_DST = obj
        refresh_model_cache_signal.send(sender=None)
        return obj

    @classmethod
    def get_loaded_model_src(cls) -> 'Language':
        """Get the currently loaded source language"""
        return cls.LOADED_SRC

    @classmethod
    def get_loaded_model_dst(cls) -> 'Language':
        """Get the currently loaded destination language"""
        return cls.LOADED_DST

    @classmethod
    def get_loaded_trie(cls) -> Trie:
        """Get the currently loaded language trie"""
        return cls.LOADED_TRIE

    @classmethod
    def unload_model_src(cls):
        """Unload the currently loaded source language if any"""
        cls.LOADED_TRIE = None
        current = cls.LOADED_SRC
        if current is None:
            return
        logger.info(f'Unloading Language SRC model: {current.iso1}')
        cls.LOADED_SRC = None
        refresh_model_cache_signal.send(sender=None)

    @classmethod
    def unload_model_dst(cls):
        """Unload the currently loaded destination language if any"""
        current = cls.LOADED_DST
        if current is None:
            return
        logger.info(f'Unloading Language DST model: {current.iso1}')
        cls.LOADED_DST = None
        refresh_model_cache_signal.send(sender=None)

class BaseModel(models.Model):
    """Mixin class for loading entrypoint models"""
    class Meta:
        abstract = True
    # This should be a dict of dicts where the key is the name of the option and the value contains:
    #  - type: The type of the option (str, int, float, bool)
    #  - default: The default value of the option (can be a callable that returns the default value)
    #  - description: A description of the option
    ALLOWED_OPTIONS = {}

    # Needed to run load tests on plugins without triggering load events
    DISABLE_LOAD_EVENTS = False

    # Map of key in model data to the related_name of the language field in the model
    CREATE_LANG_KEYS: dict[str, str] = {}

    LOADED_MODEL: 'BaseModel' = None

    entrypoint_namespace = None

    name = models.CharField(max_length=128)

    entrypoint = models.CharField(max_length=128, null=True)

    language_format = models.CharField(max_length=32, null=True)
    iso1_map = models.JSONField(null=True, blank=True)

    active = models.BooleanField(default=False)

    default_options = models.ForeignKey(
        OptionDict, on_delete=models.SET_NULL, related_name='used_by_%(class)s', null=True
        )

    load_events = models.ManyToManyField(LoadEvent, related_name='%(class)s')

    def __str__(self):
        return str(self.name)

    def __del__(self):
        try:
            self.unload()
        except NotImplementedError:
            pass

    def __getattribute__(self, name):
        res = super().__getattribute__(name)
        if name == 'DISABLE_LOAD_EVENTS':
            return res
        if not self.DISABLE_LOAD_EVENTS and name == 'load':
            def wrapped():
                _res = res()
                self.load_events.create(description=f'Loading model {self.name}')
                return _res
            return wrapped
        return res

    @classmethod
    def get_last_loaded(cls) -> 'BaseModel':
        """Get the last loaded model"""
        res = cls.objects
        res = res.filter(active=True)
        # This is not a problem with SQLite, but with postgres, the order is not guaranteed when no date is set
        res = res.annotate(count=models.Count('load_events'))
        res = res.filter(count__gt=0)
        res = res.order_by('-load_events__date')
        res = res.first()
        if res is None or len(res.load_events.all()) == 0:
            logger.debug(f'No load events found for {cls.__name__}')
            return None
        return res

    def get_lang_code(self, lang: 'Language') -> str:
        """Get the language code for a specific model"""
        if isinstance(self.iso1_map, dict) and lang.iso1 in self.iso1_map:
            return self.iso1_map[lang.iso1]
        return getattr(lang, self.language_format or 'iso1')

    @classmethod
    def from_dct(cls, _data: dict) -> 'BaseModel':
        """Create or update a model from a dictionary"""
        logger.debug(f'Creating/Updating {cls.__name__} from dict: {_data}')

        data = _data.copy()  # Avoid modifying the input dictionary
        def_opt = data.pop('default_options', {})

        if 'lang_code' in data:
            lang_code = data.pop('lang_code')
            if 'language_format' in data:
                logger.warning(f"Ignoring 'lang_code' in favor of 'language_format' for {data}")
            else:
                data['language_format'] = lang_code

        name = data.pop('name')
        langs = {}
        for key in cls.CREATE_LANG_KEYS:
            try:
                langs[key] = data.pop(key)
            except KeyError as exc:
                raise KeyError(f'Missing key {key} in data for {cls.__name__} {name}={data}') from exc
            if not isinstance(langs[key], (list, tuple)):
                raise TypeError(f'Key {key} must be a list or tuple in data for {cls.__name__}')

        data.setdefault('active', True)
        obj, _ = cls.objects.update_or_create(
            name=name,
            defaults=data,
        )

        for key, lang_list in langs.items():
            lang_objs = Language.objects.filter(iso1__in=lang_list).all()

            invalid = set(lang_list) - set(l.iso1 for l in lang_objs)
            if invalid:
                raise ValueError(f'Invalid languages {invalid} for {cls.__name__} <- {data}')

            model_key = cls.CREATE_LANG_KEYS[key]

            related: models.ManyToManyField[Language] = getattr(obj, model_key)

            new = set(lang_objs)
            prv = set(related.all())

            to_add = new - prv
            to_rmv = prv - new

            if to_rmv:
                related.remove(*to_rmv)
            if to_add:
                related.add(*to_add)

            if to_add or to_rmv:
                logger.debug(f'Updating {model_key} for OCRModel {name}:')
                to_add = [l.iso1 for l in to_add]
                to_rmv = [l.iso1 for l in to_rmv]
                logger.debug(f'  + {to_add}')
                logger.debug(f'  - {to_rmv}')

        opt_obj, created = OptionDict.objects.get_or_create(options=def_opt)
        if created:
            logger.debug(f'Created new OptionDict for {cls.__name__} {name}: {def_opt}')
        obj.default_options = opt_obj
        obj.save()

        return obj

    @classmethod
    def from_entrypoint(cls, name: str) -> 'BaseModel':
        """Get the entrypoint specific TSL model class from the entrypoint name"""
        if cls.entrypoint_namespace is None:
            raise ValueError('Cannot load base model class from entrypoint.')

        obj = cls.objects.get(name=name)
        ept = obj.entrypoint

        logger.debug(f'Loading model {name} from entrypoint {cls.entrypoint_namespace}:{ept}')
        for entrypoint in entry_points(group=cls.entrypoint_namespace, name=ept):
            new_cls = entrypoint.load()
            break
        else:
            raise ValueError(f'Missing plugin: Entrypoint "{ept}" not found.')

        return new_cls.objects.get(name=name)

    def load(self) -> None:
        """Placeholder method for loading the model. To be implemented via entrypoint"""
        raise NotImplementedError('The base model class does not implement this method.')

    def unload(self) -> None:
        """Placeholder method for unloading the model. To be implemented via entrypoint"""
        raise NotImplementedError('The base model class does not implement this method.')

    @classmethod
    def load_model(cls, model_name: str) -> 'BaseModel':
        """Load a model by name and unload the current one if needed"""
        current = cls.get_loaded_model()
        if current is not None:
            if current.name == model_name:
                return current
            current.unload()

        logger.info(f'Loading {cls.__name__} model: {model_name}')
        obj = cls.from_entrypoint(model_name)
        obj.load()

        cls.LOADED_MODEL = obj
        return obj

    @classmethod
    def unload_model(cls):
        """Unload the currently loaded model if it is this one"""
        current = cls.get_loaded_model()
        if current is None:
            return
        logger.info(f'Unloading {cls.__name__} model: {current.name}')
        current.unload()
        cls.LOADED_MODEL = None

    @classmethod
    def get_loaded_model(cls) -> 'BaseModel':
        """Get the currently loaded model"""
        return cls.LOADED_MODEL

    def deactivate(self):
        """Deactivate the model and unload it if it is currently loaded"""
        current = self.get_loaded_model()
        if self.active:
            self.active = False
            self.save()
        if current is not None and current.name == self.name:
            self.unload_model()

    def activate(self):
        """Activate the model"""
        if not self.active:
            self.active = True
            self.save()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Utils to manage ocr_translate plugins."""

# pylint: disable=redefined-outer-name,unspecified-encoding,too-many-branches

import functools
import json
import logging
import os
import platform
import shutil
import site
import subprocess
import sys
import time
from importlib import resources
from importlib.metadata import entry_points
from pathlib import Path
from threading import Lock

from django.apps import apps
from django.conf import settings

GROUPS = ['ocr_translate.box_data', 'ocr_translate.ocr_data', 'ocr_translate.tsl_data']
GENERIC_SCOPE = 'generic'
DEFAULT_OCT_BASE_DIR: Path = Path.home() / '.ocr_translate'

logger = logging.getLogger('ocr.general')

def find_site_packages(start_dir: Path) -> Path:
    """Find the site-packages directory."""
    for pth in start_dir.rglob('site-packages'):
        return pth
    return None

def reload_django_apps():
    """Reload the plugins."""
    apps.app_configs = {}
    apps.apps_ready = apps.models_ready = apps.loading = apps.ready = False
    apps.clear_cache()
    apps.populate(settings.INSTALLED_APPS)

def get_safe_name(name: str) -> str:
    """Get a safe name for the plugin."""
    return name.replace('-', '_min_')

def install_overrides_decorator(func):
    """Decorator to install a package."""
    arg_names = ['version', 'extras', 'scope', 'system']

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        args = list(args)
        cls = args.pop(0)
        if args:
            if 'name' in kwargs:
                raise TypeError('Got multiple values for argument name')
            name = args.pop(0)
        elif 'name' in kwargs:
            name = kwargs.pop('name')
        else:
            raise ValueError('Name required')

        new_args = {'name': name}
        for val,key in zip(args, arg_names):
            new_args[key] = val
        for key, val in kwargs.items():
            if key in new_args:
                raise  TypeError(f'Got multiple values for argument {key}')
            new_args[key] = val

        safe_name = get_safe_name(name).upper()
        for key in arg_names:
            over_name = f'OCT_PKG_{safe_name}_{key.upper()}'
            if over_name in os.environ:
                logger.debug(f'Overriding {key} for {name} with {os.environ[over_name]}')
                new_args[key] = os.environ[over_name]

        return func(cls, **new_args)

    return wrapped

def _pip_install(package: str, prefix: str, extras: list[str] = None, retries: int = 3):
    """Install a package with pip."""
    if extras is None:
        extras = []
    logger.info(f'Installing {package}')
    cmd = [
        'pip', 'install', package,
        '--ignore-installed',
        '--no-deps',
        '--no-build-isolation',
        f'--prefix={prefix}',
        ] + extras
    attempts = 0
    while True:
        attempts += 1
        try:
            res = subprocess.run(cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as exc:
            if attempts <= retries:
                logger.error(
                    f'Error installing {package}, retrying {attempts}/{retries}. '
                    'Waiting 5s before retrying...'
                    )
                time.sleep(5)
                continue
            logger.error(f'Error installing {package}')
            logger.error(exc.stdout.decode())
            logger.error(exc.stderr.decode())
            raise exc
        break
    logger.debug(res.stdout.decode())

class PluginManager:  # pylint: disable=too-many-instance-attributes
    """Manage the plugins."""
    _SINGLETON = None

    device: str = None
    scopes: list[str] = None

    base_dir: Path = None
    plugin_dir: Path = None
    PLUGIN_SP: dict[str, Path] = {}

    plugin_list_file: Path = None
    installed_file: Path = None

    def __new__(cls, *args, **kwargs):
        if not cls._SINGLETON:
            cls._SINGLETON = super().__new__(cls, *args, **kwargs)
        return cls._SINGLETON

    def __init__(self):
        """Initialize the plugin manager."""
        self.device = os.environ.get('DEVICE', 'cpu')
        self.scopes = [GENERIC_SCOPE, self.device]
        self.base_dir = Path(os.environ.get('OCT_BASE_DIR', DEFAULT_OCT_BASE_DIR))
        self.plugin_dir = self.base_dir / 'plugins'

        self.plugin_list_file = self.base_dir / 'plugins.json'
        self.installed_file = self.plugin_dir / 'installed.json'

        self.disabled = os.environ.get('OCT_DISABLE_PLUGINS', False)

        self.lock_plugin = Lock()
        self.lock_pkg = Lock()

        self._plugins = None
        self._plugin_data = None
        self._installed = None
        self.system = platform.system().lower()

        self.ensure_pip()
        self.initialize_scopes()

        # Ensure no previous empty or wrong site-packages dirs are present
        while (tmp_dir := find_site_packages(self.plugin_dir)) is not None:
            shutil.rmtree(tmp_dir)

    @property
    def manual_plugin_list(self) -> list[str]:
        """Get the list of manually installed plugins."""
        if self.disabled:
            return []
        modules = set()
        for grp in GROUPS:
            for ep in entry_points(group=grp):
                modules.add(ep.module)
        return list(modules)

    def ensure_pip(self):
        """Ensure pip is installed."""
        if self.disabled:
            return
        try:
            subprocess.run(['pip', '-V'], capture_output=True, check=True)
        except FileNotFoundError as exc:
            msg = ' ---- '.join([
                '!!!!!'
                'pip not found, please install a version of python https://www.python.org/downloads/'
                'Make sure to check the `Add python.exe to PATH` checkbox during installation.'
            ])
            logger.error(msg)
            raise FileNotFoundError(msg) from exc

    @property
    def plugins_data(self) -> list[dict]:
        """Get/cache the plugin data."""
        if self.disabled:
            return []
        if self._plugin_data is None:
            data_file = resources.files('ocr_translate') / 'plugins_data.json'
            if not data_file.exists():
                self._plugin_data = []
            else:
                with open(data_file) as f:
                    self._plugin_data = json.load(f)
        return self._plugin_data

    def get_plugin_data(self, name: str) -> dict:
        """Get the data for a specific plugin."""
        for plugin in self.plugins_data:
            if plugin['name'] == name:
                return plugin
        return {}

    def initialize_scopes(self):
        """Configure the scopes."""
        if self.disabled:
            return
        for scope in self.scopes:
            pth = self.plugin_dir / scope
            pth.mkdir(exist_ok=True, parents=True)
            self.PLUGIN_SP[scope] = pth
            sys.path.insert(0, pth.as_posix())
            logger.debug(f'Added {pth} to sys.path')
        sys.path_importer_cache.clear()

        site.USER_BASE = self.plugin_dir.as_posix()
        site.USER_SITE = self.PLUGIN_SP[GENERIC_SCOPE].as_posix()

    @property
    def all_plugins(self) -> list[str]:
        """Get the list of all known plugins."""
        managed = set(self.managed_plugins)
        manual = set(self.manual_plugin_list)
        if managed & manual:
            logger.warning(f'Some plugins are both manually and automatically managed: {managed & manual}')
            logger.warning('Managed ones will most likely take precedence over manually installed ones')

        return list(managed | manual)

    @property
    def managed_plugins(self) -> list[str]:
        """Get/cache the list of installed plugins."""
        if self.disabled:
            return []
        if self._plugins is None:
            self._plugins = []
            if self.plugin_list_file.exists():
                with open(self.plugin_list_file) as f:
                    self._plugins = json.load(f)
        return self._plugins

    @property
    def installed_pkgs(self) -> dict:
        """Get/cache the installed packages."""
        if self.disabled:
            return {}
        if self._installed is None:
            self._installed = {}
            if self.installed_file.exists():
                with open(self.installed_file) as f:
                    self._installed = json.load(f)

        return self._installed

    def save_plugin_list(self):
        """Save the list of installed plugins."""
        if self.disabled:
            return
        with open(self.plugin_list_file, 'w') as f:
            json.dump(self.managed_plugins, f, indent=2)

    def save_installed(self):
        """Save the installed packages."""
        if self.disabled:
            return
        with open(self.installed_file, 'w') as f:
            json.dump(self.installed_pkgs, f, indent=2)

    @install_overrides_decorator
    def install_package(
            self,
            name: str,
            version: str,
            extras: list| str = None,
            scope: str = GENERIC_SCOPE,
            system: str = '',
            ):
        """Install a package."""
        with self.lock_pkg:
            if scope not in self.scopes:
                # logging.debug(f'Skipping {name}=={version} for scope {scope}')
                logger.debug(f'Skipping {name}=={version} for scope {scope}')
                return
            if system and system.lower() != self.system:
                logger.debug(f'Skipping {name}=={version} for system {system}')
                return

            if extras is None:
                extras = []
            elif isinstance(extras, str):
                extras = list(filter(None, extras.split(' ')))
            scoped_name = f'{scope}::{name}'

            if scoped_name in self.installed_pkgs:
                if self.installed_pkgs[scoped_name]['version'] == version:
                    logger.debug(f'{scoped_name}=={version} already installed')
                    return
                old_version = self.installed_pkgs[scoped_name]['version']
                logger.debug(f'Uninstalling {scoped_name}=={old_version}')
                self._uninstall_package(scoped_name)

            ptr = {}
            self.installed_pkgs[scoped_name] = ptr
            ptr['version'] = version
            ptr['files'] = []
            ptr['dirs'] = []

            _pip_install(f'{name}=={version}', self.plugin_dir, extras)

            tmp_dir = find_site_packages(self.plugin_dir)
            if tmp_dir is None:
                raise FileNotFoundError(f'Could not find site-packages in {self.plugin_dir}')
            logger.debug(f'Package installed in {tmp_dir}')
            for src in tmp_dir.iterdir():
                if src.name.endswith('__pycache__'):
                    continue
                # shutil.move(d, PLUGIN_SP[scope])
                dst = self.PLUGIN_SP[scope] / src.name
                if src.is_file():
                    ptr['files'].append(src.name)
                    if dst.exists():
                        dst.unlink()
                    shutil.move(src, dst)
                elif src.is_dir():
                    ptr['dirs'].append(src.name)
                    if dst.exists():
                        shutil.copytree(src, dst, dirs_exist_ok=True)
                        shutil.rmtree(src)
                    else:
                        shutil.move(src, dst)

            # Make sure only one site-packages dir exists after installing a package
            shutil.rmtree(tmp_dir)

            self.save_installed()

    def _install_plugin(self, name: str):
        """Run the `install` part of the plugin."""
        if not (data := self.get_plugin_data(name)):
            raise ValueError(f'Plugin {name} not found')
        pkg = data['package']
        version = data['version']
        deps = data.get('dependencies', [])

        logger.info(f'Installing plugin {pkg}=={version} with dependencies')
        for dep in deps[::-1]:
            self.install_package(**dep)

        self.install_package(pkg, version)

    def install_plugin(self, name: str):
        """Ensure the plugin is installed."""
        with self.lock_plugin:
            self._install_plugin(name)
            if name not in self.managed_plugins:
                self.managed_plugins.append(name)
                self.save_plugin_list()
            if name not in settings.INSTALLED_APPS:
                settings.INSTALLED_APPS.append(name)
            reload_django_apps()

    def _uninstall_package(self, name: str):
        """Uninstall a package."""
        if name not in self.installed_pkgs:
            return
        logger.info(f'Uninstalling package {name}')
        scope, _ = name.split('::')
        if scope not in self.scopes:
            logger.debug(f'Skipping {name} for scope {scope}')
            return
        ptr = self.installed_pkgs.pop(name)
        scope_dir = self.PLUGIN_SP[scope]
        for pth in ptr['files']:
            (scope_dir / pth).unlink()
        for pth in ptr['dirs']:
            shutil.rmtree(scope_dir / pth)

        self.save_installed()


    def uninstall_package(self, name: str):
        """Uninstall a package thread safe."""
        with self.lock_pkg:
            self._uninstall_package(name)

    def uninstall_plugin(self, name: str):
        """Uninstall the plugin."""
        with self.lock_plugin:
            if name not in self.managed_plugins:
                return
            data = self.get_plugin_data(name)
            pkg = data['package']
            scope = data.get('scope', GENERIC_SCOPE)
            scoped_name = f'{scope}::{pkg}'
            logger.info(f'Uninstalling plugin {scoped_name} and non-shared dependencies')
            # logger.debug(f'INSTALLED: {INSTALLED}')
            if scoped_name not in self.installed_pkgs:
                raise ValueError(f'Plugin {name} not installed')

            other_deps = set()
            plugin_deps = set()
            torm_deps = set()
            for plugin in self.plugins_data:
                if plugin['name'] not in self.managed_plugins:
                    continue
                deps = set()
                for dep in plugin.get('dependencies', []):
                    deps.add(f"{dep.get('scope', GENERIC_SCOPE)}::{dep['name']}")
                if plugin['name'] == name:
                    plugin_deps |= deps
                else:
                    other_deps |= deps
            torm_deps = plugin_deps - other_deps

            for dep in torm_deps:
                self.uninstall_package(dep)
            self.uninstall_package(scoped_name)

            self.managed_plugins.remove(name)
            self.save_plugin_list()

            settings.INSTALLED_APPS.remove(name)
            reload_django_apps()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
import logging
import re
from typing import Generator, Union

from django.db import models

from .. import queues
from ..messaging import Message
from .base import BaseModel, Language, OptionDict, Text, safe_get_or_create

logger = logging.getLogger('ocr.general')

class TSLModel(BaseModel):
    """Translation models using hugging space naming convention"""
    ALLOWED_OPTIONS = {
        'ignore_chars': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': 'Characters to ignore during translation.',
        },
        'break_chars': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': (
                'Characters on which lines will be split for translation.\n'
                'The split lines are translated separately (no context between them) and then merged back together.)'
                ),
        },
        'allowed_start_end': {
            'type': str,
            'default': ('cascade', ['lang_src', 'tsl_model'], ''),
            'description': (
                'Characters allowed at the start and end of a line.\n'
                'Some models like tesseract, will detect the edges of text bubbles as text and spit out garbage '
                'at the start/end of the translated line. This option will filter out any caracter that is not '
                'allowed at the start/end of a line.'
                ),
        },
        'break_newlines': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': 'Split lines on newlines. If false the newlines will be replaced with spaces.',
        },
        'restore_missing_spaces': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': (
                'Some models will OCR text without spaces between some of the words.\n'
                'This option will attempt to restore the missing spaces by finding the shortest valid decomposition '
                'also keeping into account the word frequency.\n'
                '(Only available if a trie is loaded for the source language.)'
                ),
        },
        'restore_dash_newlines': {
            'type': bool,
            'default': ('cascade', ['lang_src', 'tsl_model'], False),
            'description': (
                'Text can be written with dash splitting a word onto a new line.\n'
                'If this option is enabled, dashes preceded by a character and followed by a newline will be removed '
                'togheter with the newline to merge the word back together.'
                ),
        },
    }
    CREATE_LANG_KEYS = {'lang_src': 'src_languages', 'lang_dst': 'dst_languages'}

    entrypoint_namespace = 'ocr_translate.tsl_models'

    src_languages = models.ManyToManyField(Language, related_name='tsl_models_src')
    dst_languages = models.ManyToManyField(Language, related_name='tsl_models_dst')

    @staticmethod
    def pre_tokenize( # pylint: disable=too-many-branches
            text: str,
            ignore_chars: str = None,
            break_chars: str = None,
            allowed_start_end: str = None,
            break_newlines: bool = False,
            restore_missing_spaces: bool = False,
            restore_dash_newlines: bool = False,
            **kwargs
            ) -> list[str]:
        """Pre-tokenize a text string.

        Args:
            text (str): Text to tokenize.
            lang (str): Language of the text.
            ignore_chars (str, optional): String of characters to ignore. Defaults to None.
            break_chars (str, optional): String of characters to break on. Defaults to None.
            allowed_start_end (str, optional): String of characters allowed at the start and end of a line.
            break_newlines (bool, optional): Whether to break on newlines. Defaults to True.
            restore_missing_spaces (bool, optional): Whether to restore missing spaces (2 word with no space between).
            restore_dash_newlines (bool, optional): Whether to restore dash-newlines (word broken with a -newline).
                Defaults to False.

        Returns:
            list[str]: List of string tokens.
        """
        if isinstance(break_newlines, str):
            break_newlines = break_newlines.lower() == 'true'
        if isinstance(restore_missing_spaces, str):
            restore_missing_spaces = restore_missing_spaces.lower() == 'true'
        if isinstance(restore_dash_newlines, str):
            restore_dash_newlines = restore_dash_newlines.lower() == 'true'
        orig_text = text
        if allowed_start_end is not None:
            rgx_start = re.compile(
                '(?x)'
                rf'^[^{allowed_start_end}]+\S?(?= )'
                '|'
                rf'^\S[^{allowed_start_end}]+(?= )'
                )

            rgx_end = re.compile(
                '(?x)'
                rf'(?<= )\S?[^{allowed_start_end}]+$'
                '|'
                rf'(?<= )[^{allowed_start_end}]+\S$'
                )

            app = []
            for split in text.split('\n'):
                split = rgx_start.sub('', split)
                split = rgx_end.sub('', split)
                app.append(split)
            text = '\n'.join(app)
        if restore_dash_newlines:
            text = re.sub(r'(?<!\n)- *\n', '', text)
        if ignore_chars:
            text = re.sub(f'[{ignore_chars}]+', '', text)
        if break_chars is None:
            break_chars = ''
        if break_newlines:
            break_chars += '\n'
        else:
            text = text.replace('\n', ' ')

        trie = Language.get_loaded_trie()
        if restore_missing_spaces and not trie is None:
            res = []
            for split in text.lower().split(' '):
                if not trie.search(split, strict=False):
                    decomposed = trie.decompose(split, min_length=1)
                    if decomposed:
                        res.append(decomposed)
                    else:
                        res.append([[split]])
                else:
                    res.append([[split]])

            # Use a list of word frequencies to determine the best split
            def sum_freq(lst: list) -> float:
                return sum(trie.get_freq(w) for w in lst) / len(lst)**4.0

            res = [' '.join(max(_, key=sum_freq)) for _ in filter(None, res)]
            text = ' '.join(res)

        break_chars = re.escape(break_chars)
        tokens = text
        if len(break_chars) > 0:
            tokens = re.split(f'[{break_chars}+]', text)

        if isinstance(tokens, str):
            tokens = [text]

        res = list(filter(None, tokens))
        logger.debug(f'Pre-tokenized "{orig_text}" to {res}')
        return res if len(res) > 0 else [' ']


    def _translate(
            self,
            tokens: list, src_lang: str, dst_lang: str, options: dict = None) -> str | list[str]:
        """PLACEHOLDER (to be implemented via entrypoint): Translate a text using a the loaded model.

        Args:
            tokens (list): list or list[list] of string tokens to be translated.
            lang_src (str): Source language.
            lang_dst (str): Destination language.
            options (dict, optional): Options for the translation. Defaults to {}.

        Raises:
            TypeError: If text is not a string or a list of strings.

        Returns:
            Union[str,list[str]]: Translated text. If text is a list, returns a list of translated strings.
        """
        # Redefine this method with the same signature as above
        # Should return a string with the translated text.
        # IMPORTANT: the main codebase treats this function as batchable:
        # The input `tokens` can be a list of strings or a list of list of strings. The output should match the input
        #   being a string or list of strings.
        # (This is used to leverage the capability of pytorch to batch inputs and outputs for faster performances,
        #   or it can also used to write a plugin for an online service by using a single request for multiple inputs
        #   using some separator that the service will leave unaltered.)
        raise NotImplementedError('The base model class does not implement this method.')

    def find_manual(self,  text_obj: 'Text', src: 'Language', dst: 'Language') -> 'TranslationRun':
        """Find a manual translation run for the given text with the specified source and destination languages.
        Args:
            text_obj (m.Text): Text object from the database to translate.
            src (m.Language): Source language object from the database.
            dst (m.Language): Destination language object from the database.

        Returns:
            m.TranslationRun: The TranslationRun object from the database.
        """
        manual_model, _ = TSLModel.objects.get_or_create(name='manual')
        params = {
            'model': manual_model,
            'lang_src': src,
            'lang_dst': dst,
            'text': text_obj,
            'options': OptionDict.objects.get(options={})
        }
        # logger.debug(f'Looking for manual TSL with params: {params}')
        return TranslationRun.objects.filter(**params).first()


    def translate(
            self,
            text_obj: 'Text', src: 'Language', dst: 'Language', options: 'OptionDict' = None,
            force: bool = False,
            block: bool = True,
            favor_manual: bool = True,
            lazy: bool = False
            ) -> Generator[Union[Message, 'Text'], None, None]:
        """High level translate call generating a TranslationRun entry.
        Args:
            text_obj (m.Text): Text object from the database to translate.
            src (m.Language): Source language object from the database.
            dst (m.Language): Destination language object from the database.
            options (m.OptionDict, optional): OptionDict object from the database. Defaults to None.
            force (bool, optional): Whether to force a new TSL run. Defaults to False.
            block (bool, optional): Whether to block until the task is complete. Defaults to True.
            favor_manual (bool, optional): Whether to favor manual translations over TSL. Defaults to True.
            lazy (bool, optional): Whether to raise an error if the TSL run is not found. Defaults to False.

        Raises:
            ValueError: If lazy and force are both True or if lazy is True and the TSL run is not found.

        Yields:
            Generator[Union[Message, m.Text], None, None]:
                If block is False, yields a Message object for the TSL run first and the resulting Text object second.
                If block is True, yields the resulting Text object.
        """
        if lazy and force:
            raise ValueError('Cannot force + lazy TSL run')
        # Check if a ManualModel is being used
        tsl_run_obj = None
        if favor_manual:
            tsl_run_obj = self.find_manual(text_obj, src, dst)
        if tsl_run_obj is None:
            options_obj = options or OptionDict.objects.get(options={})
            params = {
                'options': options_obj,
                'text': text_obj,
                'model': self,
                'lang_src': src,
                'lang_dst': dst,
            }
            tsl_run_obj = TranslationRun.objects.filter(**params).first()
        if tsl_run_obj is None or force:
            if lazy:
                raise ValueError('Value not found for lazy TSL run')
            logger.info('Running TSL')
            # Generate a unique id for a message
            id_ = (text_obj.id, self.id, options_obj.id, src.id, dst.id, options_obj.id)
            batch_id = (self.id, options_obj.id, src.id, dst.id, options_obj.id)
            lang_dct = getattr(src.default_options, 'options', {})
            model_dct =  getattr(self.default_options, 'options', {})
            opt_dct = {**lang_dct, **model_dct, **options_obj.options}

            tokens = self.pre_tokenize(text_obj.text, **opt_dct)
            new = queues.tsl_queue.put(
                id_=id_,
                batch_id=batch_id,
                handler=self._translate,
                msg={
                    'args': (
                        tokens,
                        self.get_lang_code(src),
                        self.get_lang_code(dst),
                        ),
                    'kwargs': {'options': opt_dct},
                },
            )
            if not block:
                yield new
            new = new.response()
            text_obj = safe_get_or_create(Text, text=new)
            params['result'] = text_obj
            tsl_run_obj = TranslationRun.objects.create(**params)
        else:
            if not block:
                # Both branches should have the same number of yields
                yield None
            manual = ''
            if tsl_run_obj.model.name == 'manual':
                manual = 'manual '
            logger.info(f'Reusing {manual}TSL <{tsl_run_obj.id}>')

        yield tsl_run_obj.result

class TranslationRun(models.Model):
    """Translation run on a text using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='trans_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_src')
    lang_dst = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='trans_dst')

    text = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='to_trans')
    model = models.ForeignKey(TSLModel, on_delete=models.CASCADE, related_name='tsl_runs')
    result = models.ForeignKey(Text, on_delete=models.CASCADE, related_name='from_trans')



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django routing for the ocr_translate app."""
from django.urls import path

from ocr_translate import views

app_name = 'ocr_translate'
urlpatterns = [
    path('', views.handshake, name='handshake'),
    path('set_models/', views.set_models, name='set_models'),
    path('set_lang/', views.set_lang, name='set_lang'),
    path('get_trans/', views.get_translations, name='get_trans'),
    path('run_tsl/', views.run_tsl, name='run_tsl'),
    path('run_tsl_xua', views.run_tsl_get_xunityautotrans, name='run_tsl_get_xunityautotrans'),
    path('run_ocrtsl/', views.run_ocrtsl, name='run_ocrtsl'),
    path('set_manual_translation/', views.set_manual_translation, name='set_manual_translation'),
    path('get_active_options/', views.get_active_options, name='get_active_options'),
    path('get_plugin_data/', views.get_plugin_data, name='get_plugin_data'),
    path('manage_plugins/', views.manage_plugins, name='manage_plugins'),
]



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
import logging
from typing import Generator, Union

import numpy as np
from django.db import models
from PIL.Image import Image as PILImage

from .. import queues
from ..messaging import Message
from .base import BaseModel, Language, OptionDict, Text, safe_get_or_create
from .box import BBox

logger = logging.getLogger('ocr.general')

class OCRModel(BaseModel):
    """OCR model."""
    CREATE_LANG_KEYS = {'lang': 'languages'}

    entrypoint_namespace = 'ocr_translate.ocr_models'
    # iso1 code for languages that do not use spaces to separate words
    _NO_SPACE_LANGUAGES = ['ja', 'zh', 'zht', 'lo', 'my']
    _VERTICAL_LANGS = ['ja', 'zh', 'zht', 'ko']
    SINGLE='single'
    MERGED='merged'
    OCR_MODE_CHOICES = [
        (SINGLE, 'Single'),
        (MERGED, 'Merged'),
    ]

    languages = models.ManyToManyField(Language, related_name='ocr_models')
    ocr_mode = models.CharField(max_length=32, choices=OCR_MODE_CHOICES, default=MERGED)
    tokenizer_name = models.CharField(max_length=128, null=True)
    processor_name = models.CharField(max_length=128, null=True)

    def prepare_image(
            self,
            img: PILImage, bbox: tuple[int, int, int, int] = None
            ) -> PILImage:
        """Standard operation to be performed on image before OCR. E.G color scale and crop to bbox"""
        if not isinstance(img, PILImage):
            raise TypeError(f'img should be PIL Image, but got {type(img)}')
        img = img.convert('RGB')

        if bbox:
            img = img.crop(bbox)

        return img

    @staticmethod
    def merge_single_result( # pylint: disable=too-many-locals
            lang: str,
            texts: list[str],
            bboxes_single: list['BBox'],
            bboxes_merged: list['BBox'],
            ) -> list[str]:
        """Merge the results of an OCR run on the single components of a detected bounding box.
        Will try to understand if the text is horizontal or vertical and merge accordingly.

        Args:
            lang (str): Language of the text in iso1 format.
            texts (list[str]): List of texts for each component.
            bboxes_single (list['BBox']): List of single bounding boxes.
            bboxes_merged (list['BBox']): List of merged bounding boxes.

        Returns:
            list[str]: The merged text.
        """
        grouped = {}
        # Group texts and single_bboxes by merged bounding box
        for text, bbox in zip(texts, bboxes_single):
            merged = bbox.to_merged
            ptr = grouped.setdefault(merged, [])
            ptr.append((text, *bbox.lbrt))

        sep = '' if lang in OCRModel._NO_SPACE_LANGUAGES else ' '

        res = []
        for bbox in bboxes_merged:
            single_array = np.array(grouped[bbox], dtype=object)
            l,b,r,t = bbox.lbrt
            width = r - l
            height = t - b

            # Vertical langs can also be written horizontal
            # Assume that if box.width > box.height * 1.3 then it's horizontal
            vertical = lang in OCRModel._VERTICAL_LANGS and height * 1.3 > width

            if vertical:
                # Vertical text find columns
                thr = np.average(single_array[:, 3] - single_array[:, 1]) / 1.5
                centers = (single_array[:, 1] + single_array[:, 3]) / 2
            else:
                # Horizontal text find lines
                thr = np.average(single_array[:, 4] - single_array[:, 2]) / 1.5
                centers = (single_array[:, 2] + single_array[:, 4]) / 2

            # Group lines/cols if the centers are closer than the average line width/height (vertical/non-vert)
            classifiers = np.empty(0)
            line_classifiers = []
            for cen in centers:
                if len(classifiers) == 0:
                    classifiers = np.array([cen])
                    line_classifiers.append(0)
                    continue
                # Index of classifier closest to the center
                i_min = np.argmin(np.abs(classifiers - cen))
                # If the closest classifier is closer than the threshold, reuse it
                if np.abs(classifiers[i_min] - cen) < thr:
                    line_classifiers.append(i_min)
                # Otherwise create a new classifier
                else:
                    classifiers = np.append(classifiers, cen)
                    line_classifiers.append(len(classifiers) - 1)

            line_classifiers = np.array(line_classifiers)
            # Sort lines/cols top-to-bottom or right-to-left
            i_sort = np.argsort(classifiers)[::-1 if vertical else 1]
            text_chunks = []
            # Sort lines/cols chunks left-to-right or top-to-bottom
            sort_col_index = 4 if vertical else 1
            for i in i_sort:
                ind = np.where(line_classifiers == i)[0]
                j_sort = np.argsort(single_array[ind, sort_col_index])
                text_chunks += single_array[ind[j_sort], 0].tolist()

            res.append(sep.join(text_chunks))

        return res

    def _ocr(
            self,
            img: PILImage, lang: str = None, options: dict = None
            ) -> str:
        """PLACEHOLDER (to be implemented via entrypoint): Perform OCR on an image.

        Args:
            img (Image.Image):  A Pillow image on which to perform OCR.
            lang (str, optional): The language to use for OCR. (Not every model will use this)
            bbox (tuple[int, int, int, int], optional): The bounding box of the text on the image in lbrt format.
            options (dict, optional): A dictionary of options to pass to the OCR model.

        Raises:
            TypeError: If img is not a Pillow image.

        Returns:
            str: The text extracted from the image.
        """
        # Redefine this method with the same signature as above
        # Should return a string with the result of the OCR performed on the input PILImage.
        # Unless the methods `prepare_image` or `ocr` are also being overwritten, the input image will be the
        #  result of the CROP on the original image using the bounding boxes given by the box detection model.
        raise NotImplementedError('The base model class does not implement this method.')

    def ocr(
            self,
            bbox_obj: 'BBox', lang: 'Language',  image: PILImage = None, options: 'OptionDict' = None,
            force: bool = False, block: bool = True,
            ) -> Generator[Union[Message, 'Text'], None, None]:
        """High level function to perform OCR on an image.

        Args:
            bbox_obj (m.BBox): The BBox object from the database.
            lang (m.Language): The Language object from the database.
            image (Image.Image, optional): The image on which to perform OCR. Needed if no previous OCR run exists, or
                force is True.
            options (m.OptionDict, optional): The OptionDict object from the database
                containing the options for the OCR.
            force (bool, optional): Whether to force the OCR to run again even if a previous run exists.
                Defaults to False.
            block (bool, optional): Whether to block until the task is complete. Defaults to True.

        Raises:
            ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

        Yields:
            Generator[Union[Message, m.Text], None, None]:
                If block is False, yields a Message object for the OCR run first and the resulting Text object second.
                If block is True, yields the resulting Text object.
        """
        options_obj = options
        if options_obj is None:
            options_obj = OptionDict.objects.get(options={})
        params = {
            'bbox': bbox_obj,
            'model': self,
            'lang_src': lang,
            'options': options_obj,
        }
        ocr_run_obj = OCRRun.objects.filter(**params).first()
        if ocr_run_obj is None or force:
            if image is None:
                raise ValueError('Image is required for OCR')
            logger.info('Running OCR')

            id_ = (bbox_obj.id, self.id, lang.id, options_obj.id)
            mlang = self.get_lang_code(lang)
            opt_dct = options_obj.options
            text = queues.ocr_queue.put(
                id_=id_,
                handler=self._ocr,
                msg={
                    'args': (self.prepare_image(image, bbox_obj.lbrt),),
                    'kwargs': {
                        'lang': mlang,
                        'options': opt_dct
                        },
                },
            )
            if not block:
                yield text
            text = text.response()
            if lang.iso1 in self._NO_SPACE_LANGUAGES:
                text = text.replace(' ', '')

            text_obj = safe_get_or_create(Text, text=text)
            params[f'result_{self.ocr_mode}'] = text_obj
            ocr_run_obj = OCRRun.objects.create(**params)
        else:
            if not block:
                # Both branches should have the same number of yields
                yield None
            logger.info(f'Reusing OCR <{ocr_run_obj.id}>')
            # It is possible that single has to be performed is a previous pipeline was interrupted
            text_obj = ocr_run_obj.result_merged or ocr_run_obj.result_single
            # text = ocr_run.result.text

        yield text_obj

class OCRRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='ocr_src')

    # image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_ocr')
    bbox = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='to_ocr')
    model = models.ForeignKey(OCRModel, on_delete=models.CASCADE, related_name='ocr_runs')
    result_single = models.ForeignKey(
        Text, on_delete=models.CASCADE, related_name='from_ocr_single',
        default=None, null=True
        )
    result_merged = models.ForeignKey(
        Text, on_delete=models.CASCADE, related_name='from_ocr_merged',
        default=None, null=True
        )



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Run the django migrations and start the server."""
# pylint: disable=import-outside-toplevel,too-many-statements,invalid-name,wrong-import-position
import importlib
import os
import subprocess
from pathlib import Path

import django
from django.core.management import call_command

from .. import __version__
from .. import plugin_manager as pm


def banner():
    """Print the banner."""
    print('--------------------------------------------')
    print('-                                          -')
    print(f'-     OCR_TRANSLATE version v{__version__:<13s} -')
    print('-                                          -')
    print('--------------------------------------------')

def env_default():
    """Set default environment variables."""
    os.environ['DJANGO_SETTINGS_MODULE'] = 'ocr_translate.app.settings'
    os.environ.setdefault('DJANGO_DEBUG', 'True')
    os.environ.setdefault('DJANGO_LOG_LEVEL', 'INFO')

def dir_check():
    """Check if required directories exist and create them if needed."""
    if not 'OCT_BASE_DIR' in os.environ:
        base = Path.home()  / '.ocr_translate'
        print(f'OCT_BASE_DIR not set:  Using default "{base}" as base dir for ocr_translate')
        os.environ['OCT_BASE_DIR'] = base.as_posix()
    else:
        base = Path(os.environ['OCT_BASE_DIR'])
        print(f'OCT_BASE_DIR: Using "{base.as_posix()}" as base dir for ocr_translate')
    base.mkdir(exist_ok=True, parents=True)

    if not 'DATABASE_NAME' in os.environ:
        db_path = base / 'db.sqlite3'
        print(f'DATABASE_NAME not set:  Using default "{db_path}" as database')
        os.environ['DATABASE_NAME'] = (db_path).as_posix()
        base.mkdir(exist_ok=True, parents=True)
    elif (db_name := os.environ['DATABASE_NAME']).endswith('.sqlite3'):
        Path(db_name).parent.mkdir(exist_ok=True, parents=True)
        print(f'DATABASE_NAME: Using "{db_name}" as database')

def cuda_check():
    """Check if cuda is available and set the environment variable DEVICE."""
    print('Checking for CUDA availability...')
    if 'DEVICE' in os.environ:
        device = os.environ['DEVICE'].lower()
        if device not in ['cpu', 'cuda']:
            print(f'Invalid DEVICE environment variable value `{device}`: defaulting to `cpu`')
            os.environ['DEVICE'] = 'cpu'
            device = 'cpu'
        else:
            print(f'Device set via environment variable to: `{device}`')
        if device == 'cpu':
            return
    else:
        try:
            subprocess.run(['nvidia-smi'], check=True, capture_output=True)
        except FileNotFoundError:
            pass
        else:
            print('`nvidia-smi` is available, setting DEVICE to `cuda`')
            os.environ['DEVICE'] = 'cuda'

    pmng = pm.PluginManager()  # Make sure the plugin manager is initialized to have installed plugins libs in path
    try:
        # The version of torch found depends on the current scope.
        # This check is still useful to se if a CUDA capable torch is installed but a GPU is not available.
        torch = importlib.import_module('torch')
    except ModuleNotFoundError:
        print('Torch not found: cannot check for CUDA availability explicitly. Defaulting to CPU')
        os.environ['DEVICE'] = 'cpu'
    except ImportError:
        print('Torch found but cannot be imported: probably using a newer version of python with an old plugin')
        print('installation directory. The tool will attempt to update the plugins.')
        print('In case of errors please try to manually delete the following:')
        print(' - ', pmng.plugin_dir)
        print(' - ', pmng.plugin_list_file)
    else:
        print(f'Torch found `{torch.__path__}`')
        print('Using it for CUDA availability check...')
        if not torch.cuda.is_available():
            print('TORCH: CUDA is not available, falling back to using CPU')
            os.environ['DEVICE'] = 'cpu'
        else:
            os.environ.setdefault('DEVICE', 'cuda')
            print('TORCH: CUDA is available, using GPU')
        importlib.reload(pm)  # Reload the plugin manager to have the correct DEVICE set

def init():
    """Run server initializations"""
    from .. import models as m
    from ..ocr_tsl import initializers as ini
    if m.Language.objects.count() == 0:
        print('Autocreate language entries in database...')
        ini.auto_create_languages()
    ini.ensure_plugins()
    ini.sync_models_epts()
    ini.env_var_init()

def superuser():
    """Create a superuser if it does not exist and check for default password."""
    default_su_password = 'password'
    su_name = os.environ.get('DJANGO_SUPERUSER_USERNAME', '') or 'admin'
    su_pass = os.environ.get('DJANGO_SUPERUSER_PASSWORD', '') or default_su_password
    if su_name and su_pass:
        from django.contrib.auth import get_user_model
        User = get_user_model()
        if not User.objects.filter(username=su_name).exists():
            print(f'Creating superuser "{su_name}" with password "{su_pass}"')
            User.objects.create_superuser(su_name, '', su_pass)
        else:
            print(f'Superuser "{su_name}" already exists')
            if User.objects.get(username=su_name).check_password(default_su_password):
                print(f'   password still set to the default "{default_su_password}"')

def start():
    """Start the server."""
    bind_address = os.environ.get('OCT_DJANGO_BIND_ADDRESS', '127.0.0.1')
    port = os.environ.get('OCT_DJANGO_PORT', '4000')

    try:
        importlib.import_module('gunicorn')
    except ImportError:
        print('...using django development server')
        call_command('runserver', '--noreload', f'{bind_address}:{port}')
    else:
        print('...using gunicorn')
        import getpass
        user = os.environ.get('OCT_GUNICORN_USER', getpass.getuser())
        timeout = os.environ.get('OCT_GUNICORN_TIMEOUT', '1200')
        num_workers = os.environ.get('OCT_GUNICORN_NUM_WORKERS', '1')
        subprocess.run([
            'gunicorn', 'ocr_translate.app.wsgi',
            '--user', user,
            '--bind', f'{bind_address}:{port}',
            '--timeout', timeout,
            '--workers', num_workers,
        ], check=True)

def main():
    """Run the django migrations and start the server."""
    banner()
    env_default()
    dir_check()
    print('--------------------------------------------')
    cuda_check()

    print('--------------------------------------------')
    print('Running django setup...')
    django.setup()

    print('Create database (if needed) and apply database migrations (if any)...')
    call_command('migrate')
    print('--------------------------------------------')

    superuser()
    print('--------------------------------------------')
    print('Running server initializations...')
    init()

    print('--------------------------------------------')
    print('Starting server...')
    start()

if __name__ == '__main__':
    main()



"""
WSGI config for ocr_translate project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/4.2/howto/deployment/wsgi/
"""

import os

from django.core.wsgi import get_wsgi_application

os.environ['DJANGO_SETTINGS_MODULE'] = 'ocr_translate.app.settings'

application = get_wsgi_application()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Run the django migrations and start the server."""
import importlib
import os
import subprocess
import sys

req_version = os.environ.get('OCT_VERSION', '0.7.1').lower()

def install_upgrade(upgrade=False):
    """Install or upgrade the ocr_translate package."""
    cmd = ['pip', 'install']
    if upgrade:
        print(f'Upgrading ocr_translate to {req_version}...')
        cmd.append('--upgrade')
    else:
        print(f'Installing ocr_translate {req_version}...')
    if req_version in ['latest', 'last']:
        cmd.append('django-ocr_translate')
    else:
        cmd.append(f'django-ocr_translate=={req_version}')
    subprocess.run(cmd, check=True)
    print('...done')

if os.environ.get('OCT_AUTOUPDATE', 'false').lower() in ['true', 't', '1']:
    install_upgrade(upgrade=True)

try:
    importlib.import_module('ocr_translate')
except ImportError:
    print('`ocr_translate` not found: installing django-ocr_translate...')
    install_upgrade()

try:
    from ocr_translate.scripts import \
        run  # pylint: disable=wrong-import-position
except ImportError:
    print('Please relaunch this script to use the newly installed package')
    sys.exit(1)

if __name__ == '__main__':
    run.main()



###################################################################################
# ocr_translate - a django app to perform OCR and translation of images.          #
# Copyright (C) 2023-present Davide Grassano                                      #
#                                                                                 #
# This program is free software: you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by            #
# the Free Software Foundation, either version 3 of the License.                  #
#                                                                                 #
# This program is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                   #
# GNU General Public License for more details.                                    #
#                                                                                 #
# You should have received a copy of the GNU General Public License               #
# along with this program.  If not, see {http://www.gnu.org/licenses/}.           #
#                                                                                 #
# Home: https://github.com/Crivella/ocr_translate                                 #
###################################################################################
"""Django models for the ocr_translate app."""
import logging
from typing import TypedDict

from django.db import models
from PIL.Image import Image as PILImage

from .. import queues
from .base import BaseModel, Image, Language, OptionDict

logger = logging.getLogger('ocr.general')


class BBox(models.Model):
    """Bounding box of a text in an image"""
    l = models.IntegerField(null=False)
    b = models.IntegerField(null=False)
    r = models.IntegerField(null=False)
    t = models.IntegerField(null=False)

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='bboxes')
    from_ocr_single = models.ForeignKey(
        'OCRBoxRun', on_delete=models.CASCADE, related_name='result_single',
        default=None, null=True
        )
    from_ocr_merged = models.ForeignKey(
        'OCRBoxRun', on_delete=models.CASCADE, related_name='result_merged',
        default=None, null=True
        )
    to_merged = models.ForeignKey(
        'BBox', on_delete=models.CASCADE, related_name='from_single',
        default=None, null=True
        )

    @property
    def lbrt(self):
        """Return the bounding box as a tuple of (left, bottom, right, top)"""
        return self.l, self.b, self.r, self.t

    def __str__(self):
        return f'{self.lbrt}'

class BoxDetectionResult(TypedDict):
    """Type for the result of the box detection"""
    single: list[tuple[int, int, int, int]]
    merged: tuple[int, int, int, int]


class OCRBoxModel(BaseModel):
    """OCR model for bounding boxes"""
    #pylint: disable=abstract-method
    CREATE_LANG_KEYS = {'lang': 'languages'}

    entrypoint_namespace = 'ocr_translate.box_models'

    languages = models.ManyToManyField(Language, related_name='box_models')

    def _box_detection(
            self,
            image: PILImage, options: dict = None
            ) -> list[BoxDetectionResult]:
        """PLACEHOLDER (to be implemented via entrypoint): Perform box OCR on an image.
        Returns list of bounding boxes as dicts:
            - merged: The merged BBox as a tuple[int, int, int, int]
            - single: List of BBoxed merged into the merged BBox as a tuple[int, int, int, int]

        Args:
            image (Image.Image): A Pillow image on which to perform OCR.
            options (dict, optional): A dictionary of options.

        Raises:
            NotImplementedError: The type of model specified is not implemented.

        Returns:
            list[BoxDetectionResult]: List of dictionary with key/value pairs:
              - merged: The merged BBox as a tuple[int, int, int, int]
              - single: List of BBoxed merged into the merged BBox as a tuple[int, int, int, int]
        """
        # Redefine this method with the same signature as above
        # Should return a list of `lrbt` boxes after processing the input PILImage
        raise NotImplementedError('The base model class does not implement this method.')

    def box_detection( # pylint: disable=too-many-locals
            self,
            img_obj: 'Image', lang: 'Language', image: PILImage = None,
            force: bool = False, options: 'OptionDict' = None
            ) -> tuple[list['BBox'], list['BBox']]:
        """High level function to perform box OCR on an image. Will attempt to reuse a previous run if possible.

        Args:
            img_obj (m.Image): An Image object from the database.
            lang (m.Language): A Language object from the database (not every model is gonna use this).
            image (Image.Image, optional): The Pillow image to be used for OCR if a previous result is not found.
                Defaults to None.
            force (bool, optional): If true, re-run the OCR even if a previous result is found. Defaults to False.
            options (m.OptionDict, optional): An OptionDict object from the database. Defaults to None.

        Raises:
            ValueError: ValueError is raised if at any step of the pipeline an image is required but not provided.

        Returns:
            tuple[list['BBox'], list['BBox']]: Tuple of lists of BBox objects from the database.
                First list is the single bounding boxes, second list is the merged bounding boxes.
        """
        options_obj = options or OptionDict.objects.get(options={})
        params = {
            'image': img_obj,
            'model': self,
            'options': options_obj,
            'lang_src': lang,
        }

        bbox_run = OCRBoxRun.objects.filter(**params).first()
        # Needed to rerun the OCR from <0.4.x to >=0.4.x
        # Before only merged boxes where saved, now also the single are needed
        if isinstance(bbox_run, OCRBoxRun):
            if len(bbox_run.result_single.all()) == 0 and len(bbox_run.result_merged.all()) > 0:
                bbox_run.delete()
                bbox_run = None
        if bbox_run is None or force:
            if image is None:
                raise ValueError('Image is required for BBox OCR')
            logger.info('Running BBox OCR')
            opt_dct = options_obj.options
            id_ = (img_obj.id, self.id, lang.id, options_obj.id)
            bboxes = queues.box_queue.put(
                id_=id_,
                handler=self._box_detection,
                msg={
                    'args': (image,),
                    'kwargs': {'options': opt_dct},
                },
            )
            # bboxes_single, bboxes_merged = bboxes.response()
            bboxes_list = bboxes.response()
            # Create it here to avoid having a failed entry in DB
            bbox_run = OCRBoxRun.objects.create(**params)
            for dct in bboxes_list:
                merged = dct['merged']
                l,b,r,t = merged
                bbox_merged_obj = BBox.objects.create(
                    l=l, b=b, r=r, t=t,
                    image=img_obj,
                    from_ocr_merged=bbox_run,
                    )
                for bbox in dct['single']:
                    l,b,r,t = bbox
                    BBox.objects.create(
                        l=l, b=b, r=r, t=t,
                        image=img_obj,
                        from_ocr_single=bbox_run,
                        to_merged=bbox_merged_obj,
                        )
        else:
            logger.info(f'Reusing BBox OCR <{bbox_run.id}>')

        res_single = list(bbox_run.result_single.all())
        res_merged = list(bbox_run.result_merged.all())
        logger.debug(f'BBox OCR result: {len(res_single)} single boxes')
        logger.info(f'BBox OCR result: {len(res_merged)} merged boxes')

        return res_single, res_merged


class OCRBoxRun(models.Model):
    """OCR run on an image using a specific model"""
    options = models.ForeignKey(OptionDict, on_delete=models.CASCADE, related_name='ocr_box_options')

    lang_src = models.ForeignKey(Language, on_delete=models.CASCADE, related_name='box_src')

    image = models.ForeignKey(Image, on_delete=models.CASCADE, related_name='to_box')
    model = models.ForeignKey(OCRBoxModel, on_delete=models.CASCADE, related_name='box_runs')
    # result = models.ForeignKey(BBox, on_delete=models.CASCADE, related_name='from_ocr')
